/***************************************************************************
 * FIND.C - FindFirst, FindNext implementation                             *
 * -------------------------------------------                             *
 *                                                                         *
 * VCDID - CD-I filesystem driver                                          *
 * (c) 1997 - 1999 Vahur Sinijrv                                          *
 *                                                                         *
 * You are free to use any part of this source code, provided              *
 * that credit is given to author in such cases.                           *
 ***************************************************************************/
 
#include <basedef.h>
#include <vmm.h>
#include <error.h>
#include "vcdid.h"

#pragma VxD_PAGEABLE_CODE_SEG
#pragma VxD_PAGEABLE_DATA_SEG

typedef struct {
    DIRFILE        dirfile;
    DWORD          attr_search, attr_mmatch;
    ParsedPath     *ppath;
    unsigned short *pattern;
} FINDDATA;

void CDIToW32FindData( _WIN32_FIND_DATA* pWfd, LPCDI_DIRENT pDirEnt )
{
    dos_time dtime;

    pWfd->cAlternateFileName[0] = 0;
    pWfd->nFileSizeLow          = SWAPD( pDirEnt->size );

    /* Test test test */
    pWfd->nFileSizeLow  = ((pWfd->nFileSizeLow / COOKED_SECT_SIZE) * SECT_SIZE)
                          + (pWfd->nFileSizeLow % COOKED_SECT_SIZE);

    pWfd->nFileSizeHigh = 0;

    CDIToDosTime( pDirEnt->date, &dtime );
    pWfd->ftCreationTime  = IFSMgr_DosToWin32Time( dtime );
    pWfd->ftLastWriteTime = pWfd->ftCreationTime;
}

/* 1 = success, 0 = failure */
int Find( pioreq p )
{
    _WIN32_FIND_DATA *wfd;
    FINDDATA         *fd;
    _QWORD           qw;
    DWORD            attr;
    LPCDI_DIRENT     cur;
    LPCDI_DIRENT_EXT pExt;

    wfd         = (_WIN32_FIND_DATA*) p->ir_data;        // Get find data struct
    fd          = (FINDDATA*) p->ir_fh;
    p->ir_error = ERROR_NO_MORE_FILES;

    while ( cur = DirFile_GetNextEntry( &fd->dirfile ) )
    {
        /*
         * Check attributes
         */
        wfd->dwFileAttributes = FILE_ATTRIBUTE_READONLY;
        pExt = GET_DIRENT_EXT( cur );

        if ( pExt->attributes & CDI_FILE_ATTR_DIRECTORY )
        {
             if ( !(fd->attr_search & FILE_ATTRIBUTE_DIRECTORY) ) continue;
             wfd->dwFileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
        }

        if ( cur->flags & CDI_FILE_FLAG_HIDDEN )
        {
            if ( !( fd->attr_search & FILE_ATTRIBUTE_HIDDEN ) )
                continue;
            wfd->dwFileAttributes |= FILE_ATTRIBUTE_HIDDEN;
        }

        if ( (wfd->dwFileAttributes ^ fd->attr_mmatch) & fd->attr_mmatch ) continue;

        qw = IFSMgr_BCSToUNI( wfd->cFileName,
                              cur->name,
                              cur->name_len,
                              BCS_WANSI );
        ((char *)wfd->cFileName)[ qw.ddLower ] = 0;

        if ( IFSMgr_MetaMatch( fd->pattern, wfd->cFileName, UFLG_META | UFLG_NT ) )
        {
            CDIToW32FindData( wfd, cur );
            p->ir_error = NO_ERROR;
            break;
        }
    }

    return p->ir_error;
}

/*
 *  FindDirRec - Find Directory Record, which corresponds to
 *               the directory where ppath is pointing to.
 *
 *  RETURNS:     sector number for directory record or 0 when path
 *               was not found.
 */
DWORD FindDirRec( LPCDIFS_VOLUME l, ParsedPath *ppath )
{
    PathElement       *pel, *pel_last, *pel_next;
    LPCDI_PATH_TABLE  ptbl, ptbl_last;
    unsigned          parent, current, tmp;
    unsigned short    saved_length;
    unsigned short    buf[ MAX_PATH + 1 ];
    _QWORD            qw;
    DWORD             extent;

    pel_last  = IFSLastElement( ppath );
    ptbl      = l->ptbl;
    ptbl_last = (LPCDI_PATH_TABLE)((char*)ptbl + l->ptbl_size);
    parent    = 1;
    current   = 1;
    extent    = *(DWORD*)&ptbl->extent[0];

    for( pel = ppath->pp_elements;
         pel < pel_last && extent;
         pel = IFSNextElement( pel ) )
    {
        pel_next            = IFSNextElement( pel );
        saved_length        = pel_next->pe_length;
        pel_next->pe_length = 0;                      // Emulate NUL terminated UNICODE string
        extent              = 0;
        for ( ; ptbl < ptbl_last && !extent; (char*)ptbl += sizeof( CDI_PATH_TABLE ) +
                                                           ((ptbl->name_len[0] - 1) & 0xfffffffe) + 2,
                                            current++ )
        {
            tmp = SWAPW( *(WORD*)&ptbl->parent );
            if ( tmp > parent ) break;              /* Abort search      */
            if ( tmp != parent ) continue;          /* Descend into tree */

            /* . and .. directories are out */
            if ( ptbl->name_len[0] == 1 && (unsigned char)ptbl->name[0] < 2 )
                continue;

            //
            // Convert to UNICODE & match
            //
            qw = IFSMgr_BCSToUNI( buf,
                                  ptbl->name,
                                  ptbl->name_len[0],
                                  BCS_WANSI );

            buf[ qw.ddLower >> 1 ] = 0;
            if ( IFSMgr_MetaMatch( pel->pe_unichars, buf, UFLG_NT ) ) {
                parent = current;
                extent = *(DWORD*)&ptbl->extent[0];
            }
        }
        pel_next->pe_length = saved_length;
    }
    return SWAPD( extent );
}

/*
 * Initialize directory file used in subsequent find operations
 */
int InitDirRec( pioreq p )
{
    FINDDATA       *fd;
    LPCDIFS_VOLUME l;

    fd = (FINDDATA *)p->ir_fh;
    l  = (LPCDIFS_VOLUME)p->ir_rh;

    p->ir_error = DirFile_Init( &fd->dirfile, l, fd->ppath );
    return ( p->ir_error == NO_ERROR );
}

/* 1 = success, 0 = failure */
int InitFind( pioreq p )
{
    FINDDATA       *fd;
    PathElement    *pel;

    fd = (FINDDATA*)_HeapAllocate( sizeof( FINDDATA ), HEAPSWAP );
    if ( !fd ) {
        p->ir_error = ERROR_GEN_FAILURE;
        return 0;
    }

    zeromem( &fd->dirfile, sizeof( fd->dirfile ) );

    /*
     *  save search pattern
     */
    fd->ppath   = (ParsedPath*)_HeapAllocate( IFSPathSize( p->ir_ppath ), HEAPSWAP );
    if ( !fd->ppath ) {
        p->ir_error = ERROR_GEN_FAILURE;
        _HeapFree( fd, 0 );
        return 0;
    }
    memcpy( fd->ppath, p->ir_ppath, IFSPathSize( p->ir_ppath ) );
    pel         = IFSLastElement( fd->ppath );
    fd->pattern = pel->pe_unichars;

    /*
     * Save attributes
     */
    fd->attr_search = p->ir_attr & 0xff;
    fd->attr_mmatch = (p->ir_attr >> 8) & 0xff;

    p->ir_fh = fd;
    return 1;
}

void ReturnLabel( pioreq p )
{
    _WIN32_FIND_DATA *wfd;
    LPCDIFS_VOLUME   l;

    wfd = (_WIN32_FIND_DATA*) p->ir_data;        // Get find data struct
    l   = (LPCDIFS_VOLUME)p->ir_rh;

    zeromem( wfd, sizeof( _WIN32_FIND_DATA ) );
    memcpy( wfd->cFileName, l->label, l->label_size );
    wfd->dwFileAttributes = FILE_ATTRIBUTE_LABEL;
    p->ir_error = NO_ERROR;
}

void FindClose( pioreq p )
{
    if ( p->ir_fh )
    {
        DirFile_Done( &((FINDDATA*)p->ir_fh)->dirfile );

        if ( ((FINDDATA*)p->ir_fh)->ppath )
            _HeapFree( ((FINDDATA*)p->ir_fh)->ppath, 0 );

        _HeapFree( p->ir_fh, 0 );
    }
}

DWORD CDIFS_FindFirstFile( pioreq p )
{
    p->ir_fh = NULL;

    if ( p->ir_attr & (FILE_ATTRIBUTE_LABEL << 8) )   // Must match label ?
    {
        ReturnLabel( p );
        return p->ir_error;
    }

    if ( !InitFind( p ) )
         return p->ir_error;

    if ( p->ir_attr & FILE_ATTRIBUTE_LABEL )
        ReturnLabel( p );
    else
        if ( InitDirRec( p ) ) Find( p );


    if ( p->ir_error == NO_ERROR ) {
        p->ir_hfunc->hf_read  = CDIFS_FindNextFile;
        p->ir_hfunc->hf_write = CDIFS_Fail;
        p->ir_hfunc->hf_misc  = &cdifs_hndlmisc;
    } else
        FindClose( p );

    return p->ir_error;
}

DWORD CDIFS_FindNextFile( pioreq p )
{
    if ( p->ir_fh )
    {
        if ( !((FINDDATA*)p->ir_fh)->dirfile.pBuffer )
            if ( !InitDirRec( p ) ) return p->ir_error;

        Find( p );
    }
    else p->ir_error = ERROR_NO_MORE_FILES;

    return p->ir_error;
}

DWORD CDIFS_FindClose( pioreq p )
{
    FindClose( p );
    p->ir_error = NO_ERROR;
    return NO_ERROR;
}