/***************************************************************************
 * MOUNT.C - Volume mounting & unmounting                                  *
 * --------------------------------------                                  *
 *                                                                         *
 * VCDID - CD-I filesystem driver                                          *
 * (c) 1997 - 2000 Vahur Sinijrv                                          *
 *                                                                         *
 * You are free to use any part of this source code, provided              *
 * that credit is given to author in such cases.                           *
 ***************************************************************************/
 
#include <basedef.h>
#include <vmm.h>
#include <error.h>
#include "vcdid.h"
#include "cfg.h"

#define COPYDATE( dst, src, field ) \
    memcpy( dst->field, src->field, MAX_CDI_DATETIME )

#define TODWORD( n ) (*(DWORD*)&(n))

#define IR_FSD_UNMOUNT  1

#pragma VxD_PAGEABLE_CODE_SEG
#pragma VxD_PAGEABLE_DATA_SEG

LPCDIFS_VOLUME vol_list          = NULL;
DWORD          g_Next_Volume_ID  = 1;

DWORD Mount( pioreq p );
DWORD ReadPVD( LPCDIFS_VOLUME p );

DWORD _cdecl CDIFS_MountVolume( pioreq p )
{
    int status;

    switch( p->ir_flags ) {
    case IR_FSD_MOUNT:
            status = Mount( p );
            break;
    case IR_FSD_UNLOAD:
    case IR_FSD_UNMOUNT:
            return CDIFS_Unmount( p );
    default:
            status = ERROR_VOLUME_UNRECOGNIZED;
    }
    p->ir_error = status;
    return status;
}

//---------------------------------------------------------------------
// Name: Mount
// Desc: Mount volume
//---------------------------------------------------------------------
static DWORD Mount( pioreq p )
{
    LPCDIFS_VOLUME vol;
    DCB            *dcb;
    DWORD          status;

    //
    // Is mounting enabled ?
    //
    if ( !g_config.mount_flags )
    {
        return ERROR_VOLUME_UNRECOGNIZED;
    }

    // See if we have it mounted already
    for( vol = vol_list; vol; vol = vol->next )
    {
        if ( vol->drive == p->ir_mntdrv )
        {
            return ERROR_VOLUME_EXISTS;
        }
    }

    // Get DCB for device check
    dcb = GetPhysDCB( (BYTE)p->ir_mntdrv );
    if ( !dcb )
    {
        return ERROR_VOLUME_UNRECOGNIZED;
    }

    // Check, if this drive is CD-ROM
    if ( dcb->DCB_cmn.DCB_device_type != DCB_type_cdrom )
    {
        return ERROR_VOLUME_UNRECOGNIZED;
    }

    // Check block size too
    if ( ((VRP*)dcb->DCB_cmn.DCB_vrp_ptr)->VRP_block_size != COOKED_SECT_SIZE )
    {
        return ERROR_VOLUME_UNRECOGNIZED;
    }

    //
    // Allocate CDIFS_VOLUME struct and fill with data
    //
    vol = _HeapAllocate( sizeof( CDIFS_VOLUME ), HEAPSWAP );
    if ( !vol ) return ERROR_VOLUME_UNRECOGNIZED;

    vol->drive        = p->ir_mntdrv;
    vol->dcb          = dcb;
    vol->vrp          = (VRP*)dcb->DCB_cmn.DCB_vrp_ptr;
    vol->media_handle = p->ir_volh;
    vol->pFileList    = NULL;
    vol->nFiles       = 0;

    status = ReadPVD( vol );
    if ( status != NO_ERROR )
    {
        _HeapFree( vol, 0 );
        return status;
    }

    // Chain this CDIFS_VOLUME struct
    vol->next  = vol_list;
    vol_list  = vol;

    // Fill pioreq return values
    p->ir_rh    = vol;
    p->ir_vfunc = &cdifs_volfunc;
    return NO_ERROR;
}

//---------------------------------------------------------------------
// Name: CloseOpenFiles
// Desc: Close all open files associated with volume handle
//---------------------------------------------------------------------
static void CloseOpenFiles( LPCDIFS_VOLUME pVol )
{
    struct ioreq   req;
    LPFILEINFO     pFi, pTmp;
    int            i;

    zeromem( &req, sizeof( req ) );
    req.ir_flags = CLOSE_FINAL;
    pFi          = pVol->pFileList;

    for( pFi = pVol->pFileList; pFi; pFi = pTmp )
    {
        pTmp      = pFi->pPrev;
        req.ir_rh = pFi;

        CDIFS_Close( &req );
    }
}

//---------------------------------------------------------------------
// Name: CDIFS_Unmount
// Desc: Unmount volume
//---------------------------------------------------------------------
DWORD _cdecl CDIFS_Unmount( pioreq p )
{
    LPCDIFS_VOLUME l, prev;

    l = (LPCDIFS_VOLUME)p->ir_rh;
    if (!l) {
        p->ir_error = ERROR_GEN_FAILURE;
        return ERROR_GEN_FAILURE;
    }

    IO_ReleaseCachePages( l );
    _HeapFree( l->ptbl, 0 );

    CloseOpenFiles( l );

    //
    // Delete struct from list
    //
    if ( vol_list != l ) {
        for( prev = vol_list; prev; prev = prev->next ) {
            if ( prev->next == l ) {
                prev->next = l->next;
                break;
            }
        }
    } else {
        vol_list = l->next;
    }

    _HeapFree( l, 0 );

    p->ir_error = NO_ERROR;
    return NO_ERROR;
}

//---------------------------------------------------------------------
// Name: GetLabel
// Desc: Convert media label to UNICODE
//---------------------------------------------------------------------
static int GetLabel( char *medialbl, unsigned short *unilbl, unsigned lblsize )
{
    _QWORD qw;

    lblsize = trim_right( medialbl, lblsize-1 );
    qw = IFSMgr_BCSToUNI( unilbl, medialbl, lblsize, BCS_WANSI );
    return qw.ddLower;
}

//---------------------------------------------------------------------
// Name: CopyID
// Desc: Copy an id from media to buffer. Id's on media are padded
//       with spaces. This function returns number of characters
//       in id, not counting padding spaces.
//---------------------------------------------------------------------
static DWORD CopyID( char *dst, char *src, unsigned count )
{
    int len;

    memcpy( dst, src, count );
    return trim_right( dst, count - 1 );
}

//---------------------------------------------------------------------
// Name: ReadPVD
// Desc: Read Primary Volume Descriptor
//---------------------------------------------------------------------
static DWORD ReadPVD( LPCDIFS_VOLUME p )
{
    LPCDI_PVD buf;
    int       nptbl_sect;
    USHORT    status;
    DWORD     dwSize;
    DWORD     dwStatus;

    buf = (LPCDI_PVD) _HeapAllocate( COOKED_SECT_SIZE, HEAPSWAP );
    if ( !buf )
        return ERROR_GEN_FAILURE;

    dwStatus = IO_Read( p, buf, 16, 1, IORF_BYPASS_VOLTRK );
    if ( dwStatus ) {
        _HeapFree( buf, 0 );
        return ERROR_VOLUME_HARD_ERROR;
    }

    /*
     * Check PVD values
     */
    if ( !strncmp( buf->id,
                   CDI_STANDARD_ID,
                   sizeof( CDI_STANDARD_ID ) - 1 ) ) {
    if ( !strncmp( buf->system_id,
                   CDI_SYSTEM_ID,
                   sizeof( CDI_SYSTEM_ID ) - 1 ) ) {
    }}
    else {
        _HeapFree( buf, 0 );
        return ERROR_VOLUME_UNRECOGNIZED;
    }

    /*
     * Read path table from disc
     */
    p->ptbl_size = SWAPD( CDI_INT8B( buf->path_table_size ) );
    nptbl_sect   = (p->ptbl_size + COOKED_SECT_SIZE - 1) / COOKED_SECT_SIZE;
    p->ptbl = _HeapAllocate( nptbl_sect * COOKED_SECT_SIZE, HEAPSWAP );
    if ( !p->ptbl ) {
        _HeapFree( buf, 0 );
        return ERROR_GEN_FAILURE;
    }

    dwStatus = IO_Read( p,
        p->ptbl,
        SWAPD( *(DWORD*)(&buf->type_m_path_table) ),
        nptbl_sect,
        IORF_BYPASS_VOLTRK );

    if ( dwStatus ) {
        _HeapFree( p->ptbl, 0 );
        _HeapFree( buf, 0 );
        return ERROR_VOLUME_HARD_ERROR;
    }

    //
    // Copy label
    //
    p->label_size = GetLabel( buf->volume_id, p->label, sizeof( buf->volume_id ) );
    dwSize        = SWAPD( CDI_INT8B( buf->volume_space_size ) );
    p->serial     = dwSize;

    //
    // Copy ID's
    //
    p->volume_id_length = CopyID( p->volume_id,
        buf->volume_id,
        MAX_CDI_VOL_ID
    );

    p->album_id_length = CopyID( p->album_id,
        buf->album_id,
        MAX_CDI_ALBUM_ID
    );

    p->publisher_id_length = CopyID( p->publisher_id,
        buf->publisher_id,
        MAX_CDI_PUBLISHER_ID
    );

    p->preparer_id_length = CopyID( p->preparer_id,
        buf->preparer_id,
        MAX_CDI_PREPARER_ID
    );

    p->application_id_length = CopyID( p->application_id,
        buf->application_id,
        MAX_CDI_APPLICATION_ID
    );

    //
    // Copy dates
    //
    COPYDATE( p, buf, creation_date );
    COPYDATE( p, buf, modification_date );
    COPYDATE( p, buf, expiration_date );
    COPYDATE( p, buf, effective_date );

    p->volume_set_size = SWAPD( TODWORD( buf->volume_set_size ) );
    p->volume_sequence_number = SWAPD( TODWORD( buf->volume_sequence_number ) );

    /*
     * If dwSize is bigger than this number MulDiv will overflow
     * and throw exception 0
     *
     * Calculation is simple: 0xffffffff * COOKED_SECT_SIZE / SECT_SIZE
     *
     */

    if ( dwSize >= 0xE198E51E ) p->size = 0xffff;
    else p->size = MulDiv( dwSize, SECT_SIZE, COOKED_SECT_SIZE ) >> 4;

    _HeapFree( buf, 0 );
    return NO_ERROR;
}
