/**********************************************************************
 * statuspage.c
 *
 * Author:      Vahur Sinijrv
 * Description: Status page of CDIFS cpl
 *
 * (c) 1997 - 2000 Vahur Sinijrv
 *
 * You are free to use any part of this source code, provided
 * that credit is given to author in such cases.
 *
 *********************************************************************/

#include "headers.h"
#include <assert.h>
#include "cdifsvxd.h"
#include "error.h"
#include "resource.h"

//
// Protos
//
static LRESULT CALLBACK dlgproc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

//
// Constants
//
#define NUM_DRIVES      ((unsigned)('Z' - 'A' + 1))
#define UPDATE_TIMER_MS 1000

static const TCHAR MODULE[] = _T("STATUSPAGE");

const TCHAR cdi_datemask[] =
{
    6, 7, '/', 4, 5, '/', 0, 1, 2, 3, ' ',
    8, 9, ':', 10, 11, ':', 12, 13, ':', 14, 15
};

static PROPSHEETPAGE _psp =
{
    sizeof( PROPSHEETPAGE ),
    PSP_DEFAULT,
    NULL,
    MAKEINTRESOURCE( IDD_STATUS ),
    NULL,
    NULL,
    dlgproc,
    0,
    NULL,
    NULL
};

//
// Module variables
//
static HWND _hWnd;
static HWND _hTreeView;
static HWND _hCDILogo;

// --------------------------------------------------------------------
// Name: cdiDateToString()
// Desc: Convert date from CDI format to string
// --------------------------------------------------------------------
static BOOL cdiDateToString( TCHAR* str, char* cdidate )
{
    int   i;
    TCHAR c;

    if ( cdidate[0] == '0' )
        return FALSE;

    for( i = 0; i < sizeof( cdi_datemask ); i++ )
    {
        c = cdi_datemask[ i ];

        if ( c < 16 )
        {
            str[ i ] = cdidate[ c ];
        }
        else
        {
            str[ i ] = c;
        }
    }

    str[ i ] = 0;
    return TRUE;
}

// --------------------------------------------------------------------
// Name: clearTreeView()
// Desc: Clear tree view control
// --------------------------------------------------------------------
static void clearTreeView( void )
{
    if ( TreeView_GetCount( _hTreeView ) )
    {
        TreeView_DeleteAllItems( _hTreeView );
    }
}

// --------------------------------------------------------------------
// Name: createTreeViewItem()
// Desc: Create tree view item from volume description
// --------------------------------------------------------------------
static void createTreeViewItem( PIOCS_CDIFS_VOLUME pVol )
{
    TVINSERTSTRUCT is;
    TCHAR          buf[ 200 ];
    TCHAR          datebuf[ sizeof( cdi_datemask ) + 1 ];
    
    wsprintf( buf, "%c: (%s)", pVol->drive + 'A', pVol->volume_id );
    
    memset( &is, 0, sizeof( is ) );
    is.hParent             = TVI_ROOT;
    is.hInsertAfter        = TVI_SORT;
    is.item.mask           = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
    is.item.pszText        = buf;
    is.item.lParam         = (LPARAM)pVol->drive;
    is.item.iImage         = 0;
    is.item.iSelectedImage = 0;
    is.hParent             = TreeView_InsertItem( _hTreeView, &is );

    is.hInsertAfter        = TVI_LAST;
    is.item.mask           = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
    is.item.iImage         = 1;
    is.item.iSelectedImage = 1;

    wsprintf( buf, _T("Album: %s"), pVol->album_id );
    TreeView_InsertItem( _hTreeView, &is );

    wsprintf( buf, _T("Publisher: %s"), pVol->publisher_id );
    TreeView_InsertItem( _hTreeView, &is );

    wsprintf( buf, _T("Data Preparer: %s"), pVol->preparer_id );
    TreeView_InsertItem( _hTreeView, &is );

    wsprintf( buf, _T("Application: %s"), pVol->application_id );
    TreeView_InsertItem( _hTreeView, &is );

    wsprintf( buf, _T("Volumes in album: %d"), pVol->volume_set_size );
    TreeView_InsertItem( _hTreeView, &is );

    wsprintf( buf, _T("Album set sequence #: %d"), pVol->volume_sequence_number );
    TreeView_InsertItem( _hTreeView, &is );

    is.item.lParam  = (LPARAM)pVol->nFiles;
    is.item.mask   |= TVIF_PARAM;
    wsprintf( buf, _T("Files open: %d"), pVol->nFiles );
    TreeView_InsertItem( _hTreeView, &is );

    {
        is.item.mask &= ~TVIF_PARAM;

        if ( cdiDateToString( datebuf, pVol->creation_date ) )
        {
            wsprintf( buf, _T("Creation date: %s"), datebuf );
            TreeView_InsertItem( _hTreeView, &is );
        }

        if ( cdiDateToString( datebuf, pVol->modification_date ) )
        {
            wsprintf( buf, _T("Modification date: %s"), datebuf );
            TreeView_InsertItem( _hTreeView, &is );
        }

        if ( cdiDateToString( datebuf, pVol->expiration_date ) )
        {
            wsprintf( buf, _T("Expiration date: %s"), datebuf );
            TreeView_InsertItem( _hTreeView, &is );
        }

        if ( cdiDateToString( datebuf, pVol->effective_date ) )
        {
            wsprintf( buf, _T("Effective date: %s"), datebuf );
            TreeView_InsertItem( _hTreeView, &is );
        }
    }
}

// --------------------------------------------------------------------
// Name: updateTreeView()
// Desc: Update tree view control
// --------------------------------------------------------------------
static void updateTreeView( PIOCS_CDIFS_VOLUME* drives )
{
    HTREEITEM          hItem;
    TVITEM             item;
    PIOCS_CDIFS_VOLUME pVol;
    HTREEITEM          hNextItem;
    TCHAR              buf[ 100 ];
    int                i;

    hItem = TreeView_GetRoot( _hTreeView );

    while( hItem )
    {
        //
        // Get next drive
        //
        hNextItem = TreeView_GetNextItem( _hTreeView, hItem, TVGN_NEXT );

        //
        // Get items drive id
        //
        item.mask  = TVIF_PARAM;
        item.hItem = hItem;
        TreeView_GetItem( _hTreeView, &item );
        assert( (BYTE)item.lParam < NUM_DRIVES );

        //
        // Get ptr to volume from drives array
        //
        pVol = drives[ (BYTE)item.lParam ];
        if ( !pVol )
        {
            TreeView_DeleteItem( _hTreeView, hItem );
        }
        else
        {
            //
            // Only field, which can change is open files.
            // Seek to it and change the number.
            //
            item.hItem = TreeView_GetChild( _hTreeView, hItem );
            for( i = 0; i < 6; i++ )
            {
                item.hItem = TreeView_GetNextItem( _hTreeView, item.hItem, TVGN_NEXT );
            }

            assert( item.hItem );

            item.mask = TVIF_PARAM;
            TreeView_GetItem( _hTreeView, &item );

            if ( item.lParam != (LPARAM)pVol->nFiles )
            {
                wsprintf( buf, "Files open: %d", pVol->nFiles );

                item.mask    = TVIF_TEXT | TVIF_PARAM;
                item.pszText = buf;
                item.lParam  = (LPARAM)pVol->nFiles;
                TreeView_SetItem( _hTreeView, &item );
            }

            // Mark volume as processed
            pVol->drive = NUM_DRIVES;
        }
        hItem = hNextItem;
    }
}

// --------------------------------------------------------------------
// Name: update()
// Desc: Update page
// --------------------------------------------------------------------
static void update()
{
    DWORD               n;
    PIOCS_CDIFS_VOLUME  drives[ NUM_DRIVES ];
    PIOCS_CDIFS_VOLUMES volumes;
    PIOCS_CDIFS_VOLUME  pVol;
    DWORD               i;

    n = cdifsvxd_getMountedCount();

    if ( n == 0 )
    {
        clearTreeView();
        return;
    }

    memset( &drives, 0, sizeof( drives ) );

    volumes = (PIOCS_CDIFS_VOLUMES) LocalAlloc( LPTR,
        n * sizeof( IOCS_CDIFS_VOLUME ) + sizeof( IOCS_CDIFS_VOLUMES ) );

    assert( volumes != 0 );

    volumes->dwVolumes = n;
    n = cdifsvxd_getMounted( volumes );

    if ( n == 0 )
    {
        clearTreeView();
        LocalFree( volumes );
        return;
    }

    pVol = volumes->volumes;

    for( i = 0; i < n; i++, pVol++ )
    {
        assert( pVol->drive < NUM_DRIVES );
        drives[ pVol->drive ] = volumes->volumes + i;
    }

    updateTreeView( drives );

    //
    // Check, if all drives were processed
    //
    pVol = volumes->volumes;

    for( i = 0; i < n; i++, pVol++ )
    {
        if ( pVol->drive < NUM_DRIVES )
        {
            createTreeViewItem( pVol );
        }
    }

    LocalFree( volumes );
}

// --------------------------------------------------------------------
// Name: wm_initdialog
// Desc: Initialize status page
// --------------------------------------------------------------------
static BOOL wm_initdialog()
{
    HIMAGELIST hIml;

    _hTreeView = GetDlgItem( _hWnd, IDC_VOLUMETREE );
    _hCDILogo  = GetDlgItem( _hWnd, IDC_CDILOGO );

    hIml = ImageList_LoadImage( g_hInst,
        MAKEINTRESOURCE( IDB_TREELIST ),
        16,
        0,
        CLR_DEFAULT,
        IMAGE_BITMAP,
        LR_DEFAULTCOLOR );

    if ( !hIml )
    {
        errorw( MODULE, _T("Cannot load treeview images") );
        return FALSE;
    }

    TreeView_SetImageList( _hTreeView, hIml, TVSIL_NORMAL );
    Animate_Open( _hCDILogo, MAKEINTRESOURCE( IDR_AVI_CDILOGO ) );
    update();

    return TRUE;
}

// --------------------------------------------------------------------
// Name: psn_setactive
// Desc: Property sheet SETACTIVE notification handler
// --------------------------------------------------------------------
static BOOL psn_setactive()
{
    Animate_Play( _hCDILogo, 0, -1, -1 );
    SetTimer( _hWnd, 0, UPDATE_TIMER_MS, NULL );
    SetWindowLong( _hWnd, DWL_MSGRESULT, 0 );
    return TRUE;
}

// --------------------------------------------------------------------
// Name: psn_killactive
// Desc: Property sheet KILLACTIVE notification handler
// --------------------------------------------------------------------
static BOOL psn_killactive()
{
    KillTimer( _hWnd, 0 );
    Animate_Stop( _hCDILogo );
    return TRUE;
}

// --------------------------------------------------------------------
// Name: dlgproc
// Desc: Window procedure for status page
// --------------------------------------------------------------------
static LRESULT CALLBACK dlgproc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    switch( uMsg )
    {
    case WM_INITDIALOG:
        _hWnd = hWnd;
        return wm_initdialog();

    case WM_TIMER:
        update();
        return TRUE;

    case WM_NOTIFY:
        switch (((NMHDR FAR *)lParam)->code)
        {
        case PSN_SETACTIVE:
            return psn_setactive();

        case PSN_KILLACTIVE:
            return psn_killactive();
        }
    }

    return FALSE;
}

// --------------------------------------------------------------------
// Name: statuspage_create
// Desc: Create handle to this page
// --------------------------------------------------------------------
HPROPSHEETPAGE statuspage_create()
{
    _psp.hInstance = g_hInst;
    return CreatePropertySheetPage( &_psp );
}